/*
 Apache httpd module that uses RDMA to deliver data.

 This code is made available via the Apache License v2.0 (see LICENSE)
 and is Copyright 2007
		Dennis Dalessandro (dennis@osc.edu) - Ohio Supercomputer Center
		Pete Wyckoff (pw@osc.edu) - Ohio Supercomputer Center


 */

#define _LARGEFILE64_SOURCE
#include <stddef.h>
#include "httpd.h"
#include "http_config.h"
#include "http_log.h"
#include "http_connection.h"  /* ap_hook_pre_connection */
#include "http_request.h"  /* ap_hook_insert_filter */
#include "util_filter.h"
#include "apr_buckets.h"  /* types of specific buckets */
#include "apr_strings.h"  /* apr_strtoi64 */
#include "apr_optional.h"  /* ssl_is_https */
#include "arch/unix/apr_arch_file_io.h"  /* apr_file_t internals */
#include "ap_mpm.h"  /* ap_mpm_query */

#include "openfab.h"

const char *debug_module_name = "mod_rdma";

#define IGNORE __attribute__ ((unused))

#if 0
#define DEBUG_LEVEL 2
#define debug(lvl,fmt,args...) \
    do { \
	if (lvl <= DEBUG_LEVEL) \
	    debug_info(fmt,##args); \
    } while (0)
#else
#  define debug(lvl,fmt,...) do { } while (0)
#endif

/*
 * Debug message.
 */
static void debug_info(const char *fmt, ...)
{
    va_list ap;
    fprintf(stderr, "%s: ", debug_module_name);
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    fprintf(stderr, ".\n");
    fflush(stderr);
}

/* forward declaration */
const module AP_MODULE_DECLARE_DATA rdma_module;
static AP_DECLARE_DATA ap_filter_rec_t *rdma_output_filter_handle;

/* external function in ssl module, if it exists */
APR_DECLARE_OPTIONAL_FN(int, ssl_is_https, (conn_rec *));
static APR_OPTIONAL_FN_TYPE(ssl_is_https) *my_ssl_is_https = NULL;

/*
 * Possible RDMA modes for GET.  This should go in a shared header file.
 */
enum rdma_mode {
	RDMA_MODE_NONE = 0,       /* no "RDMA:" header */
	RDMA_MODE_SERVER_WRITES,  /* server initiates writes */
	RDMA_MODE_CLIENT_READS,   /* client pulls via reads */
};

/*
 * State for a particular connection, including all possible requests
 * inside it.
 */
struct rdma_conn_context {
	info *info;
	enum rdma_mode rdma_mode;
	char *ip;
	u_int16_t port;
	u_int32_t stag;
	u_int64_t to;
	u_int32_t maxlen;
	apr_bucket_brigade *bb_headers;  /* stored headers */
	apr_bucket_brigade *bb_pinned;   /* buckets pinned for rdma ops */
	request_rec *r;  /* current request */
	int active;  /* if not true, not RDMA or request finished, ignore */
};

struct rdma_server_context {
	info *info;  /* array of info structs for all threads */
};

/*
 * The worker.c model runs a bunch of separate servers, that are forked.
 * Each one of these servers starts one listener thread and many worker
 * threads.  After the fork, each server will call this function before
 * starting any of its threads.  We open the RDMA device here.
 * Each thread gets its own info structure.
 */
static void rdma_child_init(apr_pool_t *p, server_rec *s)
{
    int  i;
    struct rdma_server_context *server_ctx;
    int thread_limit;

    debug(6, "%s() Entry", __func__);

    server_ctx = apr_pcalloc(p, sizeof(*server_ctx));

    /* initialize the info structure for each thread */
    ap_mpm_query(AP_MPMQ_HARD_LIMIT_THREADS, &thread_limit);
    server_ctx->info = apr_pcalloc(p, thread_limit * sizeof(*server_ctx->info));

    for (i = 0; i < thread_limit; i++)
	server_ctx->info[i].rnic_state = IDLE;

    /* Walk through the other server configs and set their info the same;
     * this may not be the official way to do this. */
    do {
	ap_set_module_config(s->module_config, &rdma_module, server_ctx);
	s = s->next;
    } while (s);


    debug(6, "%s() Exit", __func__);
}

/*
 * Get a pointer to ssl-internal function, if it exists.
 */
static int rdma_post_config(apr_pool_t *pconf __attribute__((unused)),
                            apr_pool_t *plog __attribute__((unused)),
                            apr_pool_t *ptemp __attribute__((unused)),
			    server_rec *s __attribute__((unused)))
{
	debug(6, "%s() Entry", __func__);
	my_ssl_is_https = APR_RETRIEVE_OPTIONAL_FN(ssl_is_https);
	debug(6, "%s() Exit", __func__);
	return OK;
}

/*
 * Clean up all state when a connection ends.  Triggered by pool cleanup
 * after connection closes.  Opposite of pre_connection_handler below, but
 * the same info structure is used throughout a child's existence.
 */
static int rdma_connection_cleanup(void *vinfo)
{
	info *info = vinfo;
	int ret;

	debug(5, "%s() Entry", __func__);

	info->rnic_state = IDLE;

	if (info->not_tcp == 0) { /* it is ordinary TCP so bail */
		debug(4, "%s() TCP so bailing out", __func__);
		return OK;
	}
	debug(4, "%s() RDMA so can continue with teardown", __func__);

	ret = rdma_deallocate_resources(info); /*free buffer and mem reg*/
	if(ret != 0){
	    debug(0,"%s() unable to deallocate resources", __func__);
	    return -1;
	}
	debug(4, "%s() buffers are freed, stag is deallocated", __func__);
    	debug(2, "%s() Cleaning up everything", __func__);
	ret = rdma_cleanup(info); /*tear down everything else*/
	if (ret != 0)
		debug(0, "%s() Unable to complete RDMA clean up", __func__);

	debug(5, "%s() Exit", __func__);
	return OK;
}

/*
 * Preconnection handler.  Called soon after the connection is accepted,
 * before any requests are read.
 */
static int rdma_pre_connection_handler(conn_rec *c,
                                       void *csd __attribute__((unused)) )
{
	struct rdma_server_context *server_ctx;
	struct rdma_conn_context *conn_ctx;
	/* this is cheating, yes */
	struct ap_sb_handle_t {
		int child_num;
		int thread_num;
	};
	struct ap_sb_handle_t *sbh = c->sbh;
	info *info;

	debug(6, "%s() Entry", __func__);

	server_ctx = ap_get_module_config(c->base_server->module_config,
				      &rdma_module);
	info = &server_ctx->info[sbh->thread_num];

	rdma_zero_info(info);

	/* remember my info structure via the per-connection structure */
	conn_ctx = apr_pcalloc(c->pool, sizeof(*conn_ctx));
	conn_ctx->info = info;
	ap_set_module_config(c->conn_config, &rdma_module, conn_ctx);
	apr_pool_cleanup_register(c->pool, info, rdma_connection_cleanup, NULL);

	/* Add the output filter on the entire connection, but look at each
	* request to see if we do anything or not.
	*/
	ap_add_output_filter_handle(rdma_output_filter_handle, NULL, NULL, c);

	debug(6, "%s() Exit", __func__);

    return OK;
}

/*
 * Send this file bucket with rdma.  Guaranteed not to be ssl.
 */
static int ship_file(apr_bucket *b, info *info, enum rdma_mode mode)
{
	apr_bucket_file *a = b->data;
	apr_file_t *fb = a->fd;
	int fd = fb->filedes;
	int ret;

	debug(5, "%s() Entry", __func__);

	if(b->length > info->maxlen) {
		debug(0, "%s() Client does not have enough buffer",
						__func__);
		return -1;
	}


	info->not_tcp = 1;

	if (mode == RDMA_MODE_SERVER_WRITES) {
		ret = rdma_write_from_file(info, fd, b->start, b->length);
		if (ret) {
			debug(0, "%s() Can't RDMA Write File", __func__);
			return ret;
		}
	}else if (mode == RDMA_MODE_CLIENT_READS) {
		ret = setup_read_buffer_file(info, fd, b->start, b->length);
		if (ret) {
			debug(0, "%s() Can't Prep Buffer for Read", __func__);
			return ret;
		}
	} else {
		debug(0, "%s() Invalid Mode Type", __func__);
		return -1;
	}

	debug(5, "%s() Exit", __func__);
	return ret;
}

/*
 * Send this heap bucket with rdma.  May be SSL, but we've already handled
 * that: just send this heap data.
 */
static int ship_heap(apr_bucket *b, info *info, enum rdma_mode mode)
{
	apr_bucket_heap *h = b->data;
	int ret;

	debug(5, "%s() Entry", __func__);

	if(b->length > info->maxlen) {
		debug(0, "%s() Client does not have enough buffer",
						__func__);
		return -1;
	}

	info->not_tcp = 1;

	if (mode == RDMA_MODE_SERVER_WRITES) {
		ret = rdma_write_from_buffer(info, h->base + b->start, b->length);
		if (ret) {
			debug(0, "%s() Can't RDMA Write buffer", __func__);
			return ret;
		}
	} else {
		debug(0, "%s() RDMA Read not supported", __func__);
		return -1;
	}

	debug(5, "%s() Exit", __func__);

	return ret;
}

/*
 * Send this transient bucket with rdma.  So far, these only appear from
 * SSL.  Convert to heap and prepend length if server-writes, then call
 * the heap routine.
 */
static int ship_transient(apr_bucket *b, info *info, enum rdma_mode mode,
                          conn_rec *c)
{
	char *bdata = b->data;

	debug(5, "%s() Entry", __func__);

	if (my_ssl_is_https && my_ssl_is_https(c)
				&& mode == RDMA_MODE_SERVER_WRITES) {
		/* allocate a bucket 4 bytes bigger and prepend the length */
		uint32_t *blen;
		uint32_t orig_len = b->length;
		b = apr_bucket_heap_make(b, bdata + b->start - 4, b->length + 4, NULL);
		if (b == NULL)
			return 1;
		blen = (uint32_t *) b->data;
		*blen = orig_len;
	} else {
		/* directly convert to heap, inlined setaside function */
		b = apr_bucket_heap_make(b, bdata + b->start, b->length, NULL);
		if (b == NULL)
			return 1;
	}

	debug(5, "%s() Exit", __func__);

	return ship_heap(b, info, mode);
}

/*
 * Poll the CQ first to get our completion, then let the headers pass
 * through to TCP, and evenually shutdown the TCP connection.
 */
static int finish_eos_server_writes(info *info, request_rec *r, ap_filter_t *f,
                                    apr_bucket_brigade *bb)
{
	int ret;

	debug(6, "%s() Entry", __func__);

	/* wait for rdma to complete */
	ret = rdma_reap_completions(info, info->cq_count);
	if (ret) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "rdma_reap_completions");
		goto out;
	}

	debug(6, "%s() Successfully got the completion passing brigade on", __func__);

	ret = ap_pass_brigade(f->next, bb);
	if (ret != APR_SUCCESS) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, ret, r, "ap_pass_brigade");
		goto out;
	}

	debug(6, "%s() Exit", __func__);

out:
	return ret;
}

/*
 * Pass through the headers, let the client do the rdma read and resend
 * to us with a finish message.
 */
static int finish_eos_client_reads(IGNORE info *info, IGNORE request_rec *r,
			IGNORE ap_filter_t *f, IGNORE apr_bucket_brigade *bb)
{
	int ret;
	debug(5, "%s() Entry", __func__);

	debug(5, "%s() Now going to poll for ACK from client", __func__);

	ret = ap_pass_brigade(f->next, bb);
	if (ret != APR_SUCCESS) {
		ap_log_rerror(APLOG_MARK, APLOG_ERR, ret, r, "ap_pass_brigade");
		goto out;
	}

	return rdma_reap_completions(info, info->cq_count);

	debug(5, "%s() Exit", __func__);
out:
	return 0;
}

/*
 * Output filter.
 */
static apr_status_t rdma_output_filter(ap_filter_t *f, apr_bucket_brigade *bb)
{
	conn_rec *c = f->c;
	struct rdma_conn_context *conn_ctx;
	request_rec *r;
	apr_bucket_brigade *bbtmp;
	apr_bucket *b, *bnext;
	info *info;
	enum rdma_mode mode;
	int ret, seen_eos = 0;

	debug(4, "%s() Entry", __func__);

	conn_ctx = ap_get_module_config(c->conn_config, &rdma_module);
	mode = conn_ctx->rdma_mode;
	info = conn_ctx->info;
	r = conn_ctx->r;

	if (!conn_ctx->active) {
		debug(4, "%s() pass through", __func__);
		goto out;
	}

	bbtmp = apr_brigade_create(r->pool, c->bucket_alloc);

	/* hold onto the headers until we have DMAed all the data */
	debug(5, "%s() Processing Buckets", __func__);
	for (b = APR_BRIGADE_FIRST(bb); b != APR_BRIGADE_SENTINEL(bb); b = bnext)
	{
		debug(4, "%s() bucket %s", __func__, b->type->name);

		if (APR_BUCKET_IS_EOS(b)) {

			debug(5, "%s() Bucket = EOS", __func__);

			/* triggers us to pass the headers through, then this eos and
			* any other metadata headers that may follow */
			if (mode == RDMA_MODE_CLIENT_READS && !(my_ssl_is_https &&
								my_ssl_is_https(c))) {
				/* modify the headers to add RDMA: information with stags
				 * on server; cannot do this earlier in the filter chain
				 * because we don't know all the data yet. */
				/* XXX: need to store all the stags somewhere. */

				apr_bucket *bh, *bhn;
				apr_bucket_heap *bh_heap;
				char s[800];
				int len;

				debug(5, "%s() Now to fill in the headers for reply",
								__func__);

				bh = APR_BRIGADE_FIRST(conn_ctx->bb_headers);
				/* find last heap buffer in headers */
				for (;;) {
					bhn = APR_BUCKET_NEXT(bh);
					if (bhn == APR_BRIGADE_SENTINEL(
								conn_ctx->bb_headers))
						break;
					if (!APR_BUCKET_IS_HEAP(bhn))
						break;
					bh = bhn;
				}
				if (!APR_BUCKET_IS_HEAP(bh)) {
					ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
						  "first header bucket not heap");
					goto out;
				}
				#ifdef __x86_64__
				snprintf(s, sizeof(s), "RDMA: stag=%d, to=%ld len=%d\r\n\r\n",
							info->rdma_stag_index,
							info->rdma_to,
							info->rdma_bufsize);
				#else
				snprintf(s, sizeof(s), "RDMA: stag=%d, to=%lld len=%d\r\n\r\n",
							info->rdma_stag_index,
							info->rdma_to,
							info->rdma_bufsize);
				#endif
				/* make sure enough bytes to add */
				len = strlen(s);
				bh_heap = bh->data;
				if (bh->length + len - 2 >= bh_heap->alloc_len) {
					ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
						"not enough room to add RDMA header");
					goto out;
				}
				/* replace the existing ending \r\n */
				strcpy(bh_heap->base + bh->length - 2, s);
				bh->length += len - 2;
			}
			APR_BRIGADE_CONCAT(bbtmp, conn_ctx->bb_headers);
			apr_brigade_destroy(conn_ctx->bb_headers);
			seen_eos = 1;
			debug(5, "%s() Found EOS", __func__);
			/* fall through to metadata case */
		}

		debug(5, "%s() Getting next bucket", __func__);
		bnext = APR_BUCKET_NEXT(b);
		APR_BUCKET_REMOVE(b);

		if (APR_BUCKET_IS_METADATA(b)) {

			debug(5, "%s() Bucket = METADATA", __func__);

			/* always ship flushes */
			if (seen_eos || APR_BUCKET_IS_FLUSH(b)) {
				debug(4, "%s() shipping metadata (%s)", __func__,
							b->type->name);
				/* add a flush bucket before the eos */
				if (!APR_BUCKET_IS_FLUSH(b))
					APR_BRIGADE_INSERT_TAIL(bbtmp,
				apr_bucket_flush_create(c->bucket_alloc));
				APR_BRIGADE_INSERT_TAIL(bbtmp, b);  /* ship now */
			} else {
				debug(4, "%s: holding metadata (%s)", __func__,
							b->type->name);
				APR_BRIGADE_INSERT_TAIL(conn_ctx->bb_headers, b);/*hold*/
			}
		} else {
			debug(5, "%s() Bucket != METADATA", __func__);

			if (!r->sent_bodyct) {
				/* hang onto all header packets */
				debug(4, "%s() holding header block (%s)", __func__,
							b->type->name);
				APR_BRIGADE_INSERT_TAIL(conn_ctx->bb_headers, b);
			} else {
				/* data, ship it with rdma */
				debug(4, "%s() rdma data (%s)", __func__, b->type->name);

				if (APR_BUCKET_IS_FILE(b)){
					debug(5, "%s() Bucket = FILE", __func__);
					ret = ship_file(b, info, mode);
				} else if (APR_BUCKET_IS_HEAP(b)) {
					debug(5, "%s() Bucket = HEAP", __func__);
					ret = ship_heap(b, info, mode);
				} else if (APR_BUCKET_IS_TRANSIENT(b)) {
					debug(5, "%s() Bucket = TRANSIENT", __func__);
					ret = ship_transient(b, info, mode, c);
				} else {
					ret = 1;
					ap_log_rerror(APLOG_MARK, APLOG_ERR, ret, r,
						  "unknown bucket type %s", b->type->name);
				}

				if (ret){
					ap_log_rerror(APLOG_MARK, APLOG_ERR, ret, r,
						  "OPERATION FAILED ON %s", b->type->name);
					goto out;
				}


				APR_BRIGADE_INSERT_TAIL(conn_ctx->bb_pinned, b);
			}
		}
	}
	debug(5, "%s() Done Processing Buckets", __func__);

	/* bb is empty now, move the ones we kept onto it and destroy our temp */
	APR_BRIGADE_CONCAT(bb, bbtmp);
	apr_brigade_destroy(bbtmp);  /* a noop */

	if (seen_eos) {
		if (mode == RDMA_MODE_SERVER_WRITES) {
			debug(5, "%s() Handle EOS for Write", __func__);
			ret = finish_eos_server_writes(info, r, f, bb);
			if(ret != 0){
				ap_log_rerror(APLOG_MARK, APLOG_ERR, ret, r,
					  "FINISH EOS FAILED");
				goto out;
			}

		} else { /* mode == RDMA_MODE_CLIENT_READS */
			debug(5, "%s() Handle EOS for Read", __func__);
			ret = finish_eos_client_reads(info, r, f, bb);
		}
		apr_brigade_destroy(conn_ctx->bb_pinned);
		conn_ctx->active = 0;
		return ret;
	}

out:
	/* pass bb to next filter (and eventually the network) */
	ret = ap_pass_brigade(f->next, bb);
	if (ret != APR_SUCCESS)
		ap_log_rerror(APLOG_MARK, APLOG_ERR, ret, r, "ap_pass_brigade");


	debug(5, "%s() Exit", __func__);

	return ret;
}

/*
 * Hook up some filters for this particular request.  Build the request
 * context for them to use, if the headers indicate an RDMA-capable client.
 */
static void rdma_insert_filter(request_rec *r)
{
	const char *rdma_header;
	conn_rec *c = r->connection;
	struct rdma_conn_context *conn_ctx, tmp_ctx;
	info *info;
	int header_ok;
	char *token;
	const char *tmp;

	debug(5, "%s() Entry", __func__);

	conn_ctx = ap_get_module_config(c->conn_config, &rdma_module);
	conn_ctx->active = 0;  /* default ignore */
	info = conn_ctx->info;

	/* check headers for rdma-capable header? */
	rdma_header = apr_table_get(r->headers_in, "RDMA");
	if (!rdma_header){
		debug(5, "%s() Start Timer for TCP", __func__);
		goto out;
	}

	debug(5, "%s() Start Timer for iWARP", __func__);


	tmp = rdma_header;
	memset(&tmp_ctx, 0, sizeof(tmp_ctx));
	tmp_ctx.rdma_mode = RDMA_MODE_NONE;

	/* get comma-delimited tokens in standard HTTP list way */
	for (;;) {
		token = ap_get_list_item(r->pool, &tmp);
		if (!token || !*token)
			break;
		if (strcasecmp(token, "client-reads") == 0)
			tmp_ctx.rdma_mode = RDMA_MODE_CLIENT_READS;
		else if (strcasecmp(token, "server-writes") == 0)
			tmp_ctx.rdma_mode = RDMA_MODE_SERVER_WRITES;
		else if (strncasecmp(token, "ip=", 3) == 0) {
			tmp_ctx.ip = apr_palloc(c->pool, strlen(token)-3+1);
			strcpy(tmp_ctx.ip, token+3);
		} else if (strncasecmp(token, "port=", 5) == 0)
			tmp_ctx.port = (int16_t) apr_strtoi64(token+5, NULL, 0);
		else if (strncasecmp(token, "stag=", 5) == 0)
			tmp_ctx.stag = (int32_t) apr_strtoi64(token+5, NULL, 0);
		else if (strncasecmp(token, "to=", 3) == 0)
			tmp_ctx.to = apr_strtoi64(token+3, NULL, 0);
		else if (strncasecmp(token, "maxlen=", 7) == 0)
			tmp_ctx.maxlen = apr_strtoi64(token+7, NULL, 0);
		else
			ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, r,
					"unknown RDMA header token %s", token);
	}

	/*
	* Verify all mandatory fields found, and if so,
	* build the context and add the filters
	*/
	header_ok = 1;
	if (tmp_ctx.rdma_mode == RDMA_MODE_NONE) {
		ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, r,
				"improper RDMA header line no mode");
		header_ok = 0;
	}
	else if (tmp_ctx.rdma_mode == RDMA_MODE_SERVER_WRITES) {
		if (tmp_ctx.ip == NULL || tmp_ctx.port == 0 || tmp_ctx.stag == 0
				|| tmp_ctx.to == 0 || tmp_ctx.maxlen == 0) {
			ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, r,
				"improper RDMA header line for server writes");
			header_ok = 0;
			debug(0, "IP %s PORT %d STAG %d TO %d MAXLEN %d",
					tmp_ctx.ip, tmp_ctx.port, tmp_ctx.stag,
					tmp_ctx.to, tmp_ctx.maxlen);
		}
	}
	else if (tmp_ctx.rdma_mode == RDMA_MODE_CLIENT_READS) {
		if (tmp_ctx.ip == NULL || tmp_ctx.port == 0
						|| tmp_ctx.maxlen == 0) {
			ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, r,
				"improper RDMA header line for client reads");
			header_ok = 0;
		}
	}

	if (!header_ok)
		goto out;

	debug(6, "%s() Headers check out OK", __func__);

	info->maxlen = tmp_ctx.maxlen;

	/* verify against existing params if not the first req */
	if (conn_ctx->rdma_mode != RDMA_MODE_NONE) {
		if (conn_ctx->rdma_mode != tmp_ctx.rdma_mode)
			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
					"RDMA mode changed across requests");
		if (strcmp(conn_ctx->ip, tmp_ctx.ip) != 0)
			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
					"RDMA IP changed across requests");
		if (conn_ctx->port != tmp_ctx.port)
			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
					"RDMA port changed across requests");
	} else {
		/* initialize a QP and connect to the client, once per
		* connection, to be reused across multiple requests */
		int ret;
		if (info->rnic_state == IDLE) {
			perfmon_timestamp_t start, stop;
			debug(5, "%s() State = Idle, assume unconnected",
							__func__);
			info->not_tcp = 0;  /*initialize this now*/
			strcpy(info->ip_address, tmp_ctx.ip);
			info->port_override = tmp_ctx.port;
			ret = rdma_init(info);
			if (ret)
				ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
								"rdma_init");
			debug(4, "%s() Back from Init attempting to connect on %d", __func__, tmp_ctx.port);
			GET_TIME(start);
			ret = rdma_do_connect(info, tmp_ctx.ip, tmp_ctx.port);
			GET_TIME(stop);

			if (ret)
				ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
								"rdma_connect");

			debug(4, "%s() Setting state to connected", __func__);
			info->rnic_state = CONNECTED;

			debug(2, ".............CONNECTION TIME = %f milliseconds",
					elapsed_wall_time(
						start, stop, MILLISECONDS),
						__func__);

		}
		else
			debug(2, "%s() state is not idle assuming rdma still connected",
							__func__);
	}

	/* copy in new settings */
	conn_ctx->rdma_mode = tmp_ctx.rdma_mode;
	conn_ctx->ip = tmp_ctx.ip;
	conn_ctx->port = tmp_ctx.port;
	conn_ctx->stag = tmp_ctx.stag;
	conn_ctx->to = tmp_ctx.to;
	conn_ctx->maxlen = tmp_ctx.maxlen;

	/* update info structure now for later use in output filter */
	info->remote_stag = conn_ctx->stag;
	info->remote_to = conn_ctx->to;

	/* init for this request only */
	conn_ctx->r = r;
	conn_ctx->bb_headers = apr_brigade_create(r->pool, c->bucket_alloc);
	conn_ctx->bb_pinned = apr_brigade_create(r->pool, c->bucket_alloc);
	conn_ctx->active = 1;

out:

	debug(6, "%s() Exit", __func__);

}

/*
 * Register all handlers.
 */
static void rdma_register_hooks(apr_pool_t *p __attribute__((unused)) )
{
    /*
     * After config files read and modules loaded.
     */
    ap_hook_post_config(rdma_post_config, NULL, NULL, APR_HOOK_MIDDLE);

    /*
     * Each separate forked server in the worker.c model calls this
     * once, before any of its threads (listener and workers) are spawned
     */
    ap_hook_child_init(rdma_child_init, NULL, NULL, APR_HOOK_MIDDLE);

    /* notices each new connection */
    ap_hook_pre_connection(rdma_pre_connection_handler, NULL, NULL,
                           APR_HOOK_LAST);

    /* glues our filters onto each request */
    ap_hook_insert_filter(rdma_insert_filter, NULL, NULL, APR_HOOK_MIDDLE);

    /* construct filter handle(s) */

    /* If >= AF_FTYPE_CONNECTION, will hang on c->output_filters, but
     * be alloced from r->pool if r != NULL.  Seems like an apache bug.
     * But we must be after SSL to get encrypted data stream.
     */
    rdma_output_filter_handle = ap_register_output_filter("RDMA",
        rdma_output_filter, NULL, AP_FTYPE_NETWORK-1);
}


#if 0
/*no need to do any of this code but should keep it around
in case we want to do fancier things later this shows us
how to get things going*/

struct rdma_config {
    int chunk;
};

/*
 * Config file handler
 */
static void *rdma_create_server_config(apr_pool_t *p,
                                       server_rec *s __attribute__((unused)) )
{
    struct rdma_config *cfg;

    cfg = apr_palloc(p, sizeof(*cfg));
    cfg->chunk = 8192;
    return cfg;
}

static const char *rdma_config_set_chunk(cmd_parms *parms,
                                         void *mconfig __attribute__((unused)),
                                         const char *w)
{
    struct rdma_config *cfg;

    cfg = ap_get_module_config(parms->server->module_config, &rdma_module);
    cfg->chunk = atoi(w);
    return NULL;
}

static const command_rec rdma_config_cmds[] = {
    AP_INIT_TAKE1("RDMA_Chunk", rdma_config_set_chunk, NULL, RSRC_CONF,
                  "RDMA_Chunk <int> -- size of data transfers"),
    AP_INIT_NO_ARGS(NULL, NULL, NULL, 0, NULL),
};

/*
 * Interface to httpd core.
 */
const module AP_MODULE_DECLARE_DATA rdma_module =
{
    STANDARD20_MODULE_STUFF,
    .create_dir_config = NULL,
    .merge_dir_config = NULL,
    .create_server_config = rdma_create_server_config,
    .merge_server_config = NULL,
    .cmds = rdma_config_cmds,
    .register_hooks = rdma_register_hooks,
};
#else
const module AP_MODULE_DECLARE_DATA rdma_module =
{
    STANDARD20_MODULE_STUFF,
    .create_dir_config = NULL,
    .merge_dir_config = NULL,
    .create_server_config = NULL,
    .merge_server_config = NULL,
    .cmds = NULL,
    .register_hooks = rdma_register_hooks,
};

#endif
